import struct
from typing import Optional, Union


class HDPath:
    """
    A class representing an HD path. This class is the base class
    for both account specific HD paths (:class:`~ape_ledger.hdpath.HDAccountPath`)
    as well as the derivation HD path class :class:`~ape_ledger.hdpath.HDBasePath`.
    """

    def __init__(self, path: Union[str, "HDBasePath"]):
        if not isinstance(path, str) and hasattr(path, "path"):
            # NOTE: Using getattr for mypy
            path_str = getattr(path, "path")
        else:
            path_str = path

        path_str = path_str.rstrip("/")
        if not path_str.startswith("m/"):
            raise ValueError("HD path must begin with m/")

        self.path = path_str

    def __str__(self):
        return self.path


class HDAccountPath(HDPath):
    """
    An HD path where the account node is set.
    """

    def as_bytes(self):
        """
        Convert ``self.path`` to the Ledger bytes format.

        Ledger expect the following bytes input:
          1. Number of BIP 32 derivations to perform (max 10)
          2. First derivation index (big endian)
          3. Second derivation index (big endian)

        References:
        - https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki
        - https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki
        """
        elements = self.path.split("/")[1:]
        depth = len(elements)

        # Number of BIP 32 derivations to perform (max 10)
        result = bytes([depth])

        for derivation_index in elements:
            # For each derivation index in the path check if it is hardened
            hardened = "'" in derivation_index
            index = int(derivation_index.strip("'"))

            if hardened:
                # See bip32 spec for hardened derivation spec
                index = 0x80000000 | index

            # Append index to result as a big-endian (>) unsigned int (I)
            result += struct.pack(">I", index)

        return result


class HDBasePath(HDPath):
    """
    A derivation HD path useful for creating objects of type
    :class:`~ape_ledger.hdpath.HDAccountPath`.
    """

    def __init__(self, base_path: Optional[Union[str, "HDBasePath"]] = None):
        base_path = base_path or "m/44'/60'/{x}'/0/0"
        if not isinstance(base_path, str) and hasattr(base_path, "path"):
            base_path_str = base_path.path
        else:
            base_path_str = base_path

        base_path_str = base_path_str.rstrip("/")
        base_path_str = base_path_str if "{x}" in base_path_str else f"{base_path_str}/{{x}}"
        super().__init__(base_path_str)

    def get_account_path(self, account_id) -> HDAccountPath:
        return HDAccountPath(self.path.format(x=str(account_id)))


__all__ = ["HDAccountPath", "HDBasePath"]



import json
import shutil
from pathlib import Path
from tempfile import mkdtemp

import ape
import pytest
from click.testing import CliRunner
from eth_account.messages import encode_defunct
from eth_pydantic_types import HexBytes

TEST_ALIAS = "TestAlias"
TEST_HD_PATH = "m/44'/60'/{x}'/0/0"


# NOTE: Ensure that we don't use local paths for these
DATA_FOLDER = Path(mkdtemp()).resolve()
ape.config.DATA_FOLDER = DATA_FOLDER


@pytest.fixture(autouse=True, scope="session")
def clean_data_folder():
    yield  # Run all tests
    shutil.rmtree(DATA_FOLDER, ignore_errors=True)


@pytest.fixture
def hd_path():
    return TEST_HD_PATH


@pytest.fixture
def alias():
    return TEST_ALIAS


@pytest.fixture
def test_accounts():
    return ape.accounts.test_accounts


@pytest.fixture
def account_addresses(test_accounts):
    return [a.address for a in test_accounts]


@pytest.fixture
def account_0(test_accounts):
    return test_accounts[0]


@pytest.fixture
def account_1(test_accounts):
    return test_accounts[0]


@pytest.fixture
def address(account_addresses):
    return account_addresses[0]


@pytest.fixture(autouse=True)
def connection():
    with ape.networks.ethereum.local.use_provider("test") as provider:
        yield provider


@pytest.fixture
def msg_signature(account_0):
    msg = encode_defunct(text="__TEST_MESSAGE__")
    sig = account_0.sign_message(msg)
    return (
        sig.v,
        int(HexBytes(sig.r).hex(), 16),
        int(HexBytes(sig.s).hex(), 16),
    )


@pytest.fixture
def receipt(account_0, account_1):
    return account_0.transfer(account_1, "1 gwei")


@pytest.fixture
def tx_signature(receipt):
    return (
        receipt.signature.v,
        int(HexBytes(receipt.signature.r).hex(), 16),
        int(HexBytes(receipt.signature.s).hex(), 16),
    )


@pytest.fixture(autouse=True)
def mock_device(mocker, hd_path, account_addresses, msg_signature, tx_signature):
    device = mocker.MagicMock()
    device._account = hd_path
    device.get_address.side_effect = lambda *args, **kwargs: (
        account_addresses[args[0]] if args else account_addresses[0]
    )
    device.sign_message.side_effect = lambda *args, **kwargs: msg_signature
    device.sign_typed_data.side_effect = lambda *args, **kwargs: msg_signature
    device.sign_transaction.side_effect = lambda *args, **kwargs: tx_signature
    return device


@pytest.fixture
def mock_container(mocker):
    return mocker.MagicMock()


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def assert_account(address):
    def fn(account_path: Path, expected_address=None, expected_hdpath="m/44'/60'/0'/0/0"):
        expected_address = expected_address or address
        assert account_path.is_file(), "Account file missing."
        account_data = json.loads(account_path.read_text())
        assert account_data["address"] == expected_address
        assert account_data["hdpath"] == expected_hdpath

    return fn


@pytest.fixture
def device_factory(mocker, mock_device):
    def fn(module):
        patch = mocker.patch(f"ape_ledger.{module}.get_device")
        patch.return_value = mock_device

    return fn



import json
from typing import TYPE_CHECKING, Optional, cast

import pytest
from ape import networks
from ape.utils import create_tempdir
from ape_ethereum.ecosystem import DynamicFeeTransaction, StaticFeeTransaction
from eip712.messages import EIP712Message, EIP712Type
from eth_account.messages import SignableMessage
from eth_pydantic_types import HexBytes

from ape_ledger.accounts import AccountContainer, LedgerAccount
from ape_ledger.exceptions import LedgerSigningError

if TYPE_CHECKING:
    from ape.api import TransactionAPI
    from ape.types import AddressType


@pytest.fixture(autouse=True)
def patch_device(device_factory):
    return device_factory("accounts")


class Person(EIP712Type):
    name: "string"  # type: ignore # noqa: F821
    wallet: "address"  # type: ignore # noqa: F821


class Mail(EIP712Message):
    _chainId_: "uint256" = 1  # type: ignore # noqa: F821
    _name_: "string" = "Ether Mail"  # type: ignore # noqa: F821
    _verifyingContract_: "address" = "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC"  # type: ignore # noqa: F821 E501
    _version_: "string" = "1"  # type: ignore # noqa: F821

    sender: Person
    receiver: Person


ALICE_ADDRESS = "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826"
TEST_SENDER = Person(name="Alice", wallet=ALICE_ADDRESS)  # type: ignore
BOB_ADDRESS = "0xB0B0b0b0b0b0B000000000000000000000000000"
TEST_RECEIVER = Person("Bob", BOB_ADDRESS)  # type: ignore
TEST_TYPED_MESSAGE = Mail(sender=TEST_SENDER, receiver=TEST_RECEIVER)  # type: ignore
TEST_TXN_DATA = b"""`\x80`@R4\x80\x15a\x00\x10W`\x00\x80\xfd[P`\x00\x80T`\x01`\x01`\xa0\x1b\x03\x19\x163\x17\x90U`\x03\x80T`\xff\x19\x16`\x01\x17\x90Ua\x04\xa8\x80a\x00?`\x009`\x00\xf3\xfe`\x80`@R`\x046\x10a\x00pW`\x005`\xe0\x1c\x80c>G\xd6\xf3\x11a\x00NW\x80c>G\xd6\xf3\x14a\x00\xd4W\x80c\x8d\xa5\xcb[\x14a\x01\x19W\x80c\xb6\rB\x88\x14a\x01JW\x80c\xdc\r=\xff\x14a\x01RWa\x00pV[\x80c\x12)\xdc\x9e\x14a\x00uW\x80c#\x8d\xaf\xe0\x14a\x00\xa3W\x80c<\xcf\xd6\x0b\x14a\x00\xccW[`\x00\x80\xfd[4\x80\x15a\x00\x81W`\x00\x80\xfd[Pa\x00\xa1`\x04\x806\x03` \x81\x10\x15a\x00\x98W`\x00\x80\xfd[P5\x15\x15a\x01|V[\x00[4\x80\x15a\x00\xafW`\x00\x80\xfd[Pa\x00\xb8a\x01\xdcV[`@\x80Q\x91\x15\x15\x82RQ\x90\x81\x90\x03` \x01\x90\xf3[a\x00\xa1a\x01\xe5V[4\x80\x15a\x00\xe0W`\x00\x80\xfd[Pa\x01\x07`\x04\x806\x03` \x81\x10\x15a\x00\xf7W`\x00\x80\xfd[P5`\x01`\x01`\xa0\x1b\x03\x16a\x02\xddV[`@\x80Q\x91\x82RQ\x90\x81\x90\x03` \x01\x90\xf3[4\x80\x15a\x01%W`\x00\x80\xfd[Pa\x01.a\x02\xefV[`@\x80Q`\x01`\x01`\xa0\x1b\x03\x90\x92\x16\x82RQ\x90\x81\x90\x03` \x01\x90\xf3[a\x00\xa1a\x02\xfeV[4\x80\x15a\x01^W`\x00\x80\xfd[Pa\x01.`\x04\x806\x03` \x81\x10\x15a\x01uW`\x00\x80\xfd[P5a\x03\xa4V[`\x00T`\x01`\x01`\xa0\x1b\x03\x163\x14a\x01\xc9W`@\x80QbF\x1b\xcd`\xe5\x1b\x81R` `\x04\x82\x01R`\x0b`$\x82\x01Rj\x08X]]\x1a\x1b\xdc\x9a^\x99Y`\xaa\x1b`D\x82\x01R\x90Q\x90\x81\x90\x03`d\x01\x90\xfd[`\x03\x80T`\xff\x19\x16\x91\x15\x15\x91\x90\x91\x17\x90UV[`\x03T`\xff\x16\x81V[`\x00T`\x01`\x01`\xa0\x1b\x03\x163\x14a\x022W`@\x80QbF\x1b\xcd`\xe5\x1b\x81R` `\x04\x82\x01R`\x0b`$\x82\x01Rj\x08X]]\x1a\x1b\xdc\x9a^\x99Y`\xaa\x1b`D\x82\x01R\x90Q\x90\x81\x90\x03`d\x01\x90\xfd[`\x03T`\xff\x16a\x02AW`\x00\x80\xfd[`@Q3\x90G\x80\x15a\x08\xfc\x02\x91`\x00\x81\x81\x81\x85\x88\x88\xf1\x93PPPP\x15\x80\x15a\x02mW=`\x00\x80>=`\x00\xfd[P`\x00[`\x02T\x81\x10\x15a\x02\xbcW`\x00`\x02\x82\x81T\x81\x10a\x02\x8aW\xfe[`\x00\x91\x82R` \x80\x83 \x90\x91\x01T`\x01`\x01`\xa0\x1b\x03\x16\x82R`\x01\x90\x81\x90R`@\x82 \x91\x90\x91U\x91\x90\x91\x01\x90Pa\x02qV[P`@\x80Q`\x00\x81R` \x81\x01\x91\x82\x90RQa\x02\xda\x91`\x02\x91a\x03\xcbV[PV[`\x01` R`\x00\x90\x81R`@\x90 T\x81V[`\x00T`\x01`\x01`\xa0\x1b\x03\x16\x81V[`\x03T`\xff\x16a\x03\rW`\x00\x80\xfd[`\x004\x11a\x03LW`@QbF\x1b\xcd`\xe5\x1b\x81R`\x04\x01\x80\x80` \x01\x82\x81\x03\x82R`#\x81R` \x01\x80a\x04P`#\x919`@\x01\x91PP`@Q\x80\x91\x03\x90\xfd[3`\x00\x81\x81R`\x01` \x81\x90R`@\x82 \x80T4\x01\x90U`\x02\x80T\x91\x82\x01\x81U\x90\x91R\x7f@W\x87\xfa\x12\xa8#\xe0\xf2\xb7c\x1c\xc4\x1b;\xa8\x82\x8b3!\xca\x81\x11\x11\xfau\xcd:\xa3\xbbZ\xce\x01\x80T`\x01`\x01`\xa0\x1b\x03\x19\x16\x90\x91\x17\x90UV[`\x02\x81\x81T\x81\x10a\x03\xb1W\xfe[`\x00\x91\x82R` \x90\x91 \x01T`\x01`\x01`\xa0\x1b\x03\x16\x90P\x81V[\x82\x80T\x82\x82U\x90`\x00R` `\x00 \x90\x81\x01\x92\x82\x15a\x04 W\x91` \x02\x82\x01[\x82\x81\x11\x15a\x04 W\x82Q\x82T`\x01`\x01`\xa0\x1b\x03\x19\x16`\x01`\x01`\xa0\x1b\x03\x90\x91\x16\x17\x82U` \x90\x92\x01\x91`\x01\x90\x91\x01\x90a\x03\xebV[Pa\x04,\x92\x91Pa\x040V[P\x90V[[\x80\x82\x11\x15a\x04,W\x80T`\x01`\x01`\xa0\x1b\x03\x19\x16\x81U`\x01\x01a\x041V\xfeFund amount must be greater than 0.\xa2dipfsX"\x12 \\.\xe1\xb9\xbd\xde\x0b.`io)\xf2\xb6\xf1\xd5\xce\x1d\xd7_\x8b\xfd\xf6\xfbT\x14#\x1a\x12\xcf\xa7\xffdsolcC\x00\x06\x0c\x003"""  # noqa: E501


def build_transaction(
    txn: "TransactionAPI", receiver: Optional["AddressType"] = None
) -> "TransactionAPI":
    txn.chain_id = 579875
    txn.nonce = 0
    txn.gas_limit = 2
    txn.value = 10000000000
    txn.data = HexBytes(TEST_TXN_DATA)

    if receiver:
        txn.receiver = receiver

    # Values not part of RLP
    txn.sender = TEST_SENDER.wallet

    return txn


def create_static_fee_txn(
    receiver: Optional["AddressType"] = None,
) -> StaticFeeTransaction:
    txn = StaticFeeTransaction()
    txn = cast(StaticFeeTransaction, build_transaction(txn, receiver=receiver))
    txn.gas_price = 1
    return txn


def create_dynamic_fee_txn(
    receiver: Optional["AddressType"] = None,
) -> DynamicFeeTransaction:
    txn = DynamicFeeTransaction()
    txn = cast(DynamicFeeTransaction, build_transaction(txn, receiver=receiver))
    txn.max_fee = 300000000
    txn.max_priority_fee = 10000000
    return txn


TEST_STATIC_FEE_TXN = create_static_fee_txn()
TEST_STATIC_FEE_TXN_WITH_RECEIVER = create_static_fee_txn(receiver=TEST_RECEIVER.wallet)
TEST_DYNAMIC_FEE_TXN = create_dynamic_fee_txn()
TEST_DYNAMIC_FEE_TXN_WITH_RECEIVER = create_dynamic_fee_txn(receiver=TEST_RECEIVER.wallet)


@pytest.fixture
def create_account(address):
    def fn(account_path, hd_path):
        with open(account_path, "w") as account_file:
            account_data = {"address": address, "hdpath": hd_path}
            account_file.writelines(json.dumps(account_data))

    return fn


@pytest.fixture(autouse=True)
def isolated_file_system(runner):
    with runner.isolated_filesystem():
        yield


@pytest.fixture
def account(mock_container, create_account, hd_path):
    with create_tempdir() as temp_dir:
        path = temp_dir / "account.json"
        create_account(path, hd_path)
        with networks.ethereum.local.use_provider("test"):
            yield LedgerAccount(name=mock_container, account_file_path=path)


class TestAccountContainer:
    def test_save_account(self, mock_container, alias, address, hd_path, assert_account):
        container = AccountContainer(account_type=LedgerAccount)
        container.save_account(alias, address, hd_path)
        temp_dir = container.config_manager.DATA_FOLDER
        account_path = temp_dir / "ledger" / f"{alias}.json"
        assert_account(account_path, expected_hdpath=hd_path)


class TestLedgerAccount:
    def test_address_returns_address_from_file(self, account, address):
        assert account.address.lower() == address.lower()

    def test_hdpath_returns_address_from_file(self, account, hd_path):
        assert account.hdpath.path == hd_path

    def test_sign_message_personal(self, account, capsys, mock_device, msg_signature):
        message = SignableMessage(
            version=b"E", header=b"thereum Signed Message:\n6", body=b"I\xe2\x99\xa5SF"
        )
        v, r, s = account.sign_message(message)
        assert (v, int(r.hex(), 16), int(s.hex(), 16)) == msg_signature
        mock_device.sign_message.assert_called_once_with(message.body)
        output = capsys.readouterr()
        assert str(message) in output.out
        assert "Please follow the prompts on your device." in output.out

    def test_sign_message_typed(self, account, capsys, msg_signature):
        message = TEST_TYPED_MESSAGE.signable_message
        v, r, s = account.sign_message(message)
        assert (v, int(r.hex(), 16), int(s.hex(), 16)) == msg_signature
        output = capsys.readouterr()
        assert repr(message).replace("\n", "") in output.out.replace("\n", "")
        assert "Please follow the prompts on your device." in output.out

    def test_sign_message_unsupported(self, account, capsys):
        unsupported_version = b"X"
        message = SignableMessage(
            version=unsupported_version,
            header=b"thereum Signed Message:\n6",
            body=b"I\xe2\x99\xa5SF",
        )
        version_str = unsupported_version.decode("utf8")
        expected = rf"Unsupported message-signing specification, \(version={version_str}\)\."
        with pytest.raises(LedgerSigningError, match=expected):
            account.sign_message(message)

        output = capsys.readouterr()
        assert str(message) not in output.out
        assert "Please follow the prompts on your device." not in output.out

    @pytest.mark.parametrize(
        "txn",
        (
            TEST_STATIC_FEE_TXN,
            TEST_STATIC_FEE_TXN_WITH_RECEIVER,
            TEST_DYNAMIC_FEE_TXN,
            TEST_DYNAMIC_FEE_TXN_WITH_RECEIVER,
        ),
    )
    def test_sign_transaction(self, txn, mock_device, account, capsys, tx_signature):
        actual = account.sign_transaction(txn)
        v, r, s = actual.signature
        assert (v, int(r.hex(), 16), int(s.hex(), 16)) == tx_signature
        output = capsys.readouterr()
        assert str(txn) in output.out
        assert "Please follow the prompts on your device." in output.out



from typing import TYPE_CHECKING, Any, Optional, Union

import click
from ape.cli import PromptChoice

if TYPE_CHECKING:
    from click import Context, Parameter

    from ape_ledger.client import LedgerDeviceClient
    from ape_ledger.hdpath import HDAccountPath, HDBasePath


class AddressPromptChoice(PromptChoice):
    """
    A class for handling prompting the user for an address selection.
    """

    DEFAULT_PAGE_SIZE = 5

    def __init__(
        self,
        hd_path: Union["HDBasePath", str],
        index_offset: int = 0,
        page_size: int = DEFAULT_PAGE_SIZE,
    ):
        from ape_ledger.hdpath import HDBasePath

        if isinstance(hd_path, str):
            hd_path = HDBasePath(base_path=hd_path)

        self._hd_root_path = hd_path
        self._index_offset = index_offset
        self._page_size = page_size
        self._choice_index: Optional[int] = None

        # Must call ``_load_choices()`` to set address choices
        super().__init__([])

    @property
    def _is_incremented(self) -> bool:
        """Returns ``True`` if the user has paged past the first page."""
        return (self._index_offset + self._page_size) > self._page_size

    @property
    def _prompt_message(self) -> str:
        return (
            "Please choose the address you would like to add,\n\t"
            f"or type 'n' for the next {self._page_size} entries"
        )

    def convert(
        self, value: Any, param: Optional["Parameter"], ctx: Optional["Context"]
    ) -> Optional[str]:
        """Convert the user selection to a choice or increment /decrement
        if they input ``n`` or ``p``."""
        if self._page_from_choice(value):
            # Don't select an address yet if user paged.
            return None

        address = super().convert(value, param, ctx)
        address_index = self.choices.index(address)
        self._choice_index = self._choice_index if address_index is None else address_index
        return address

    def get_user_selected_account(self) -> tuple[str, "HDAccountPath"]:
        """Returns the selected address from the user along with the HD path.
        The user is able to page using special characters ``n`` and ``p``.
        """
        address = None
        while address is None:
            self._load_choices()
            self.print_choices()

            address = self._get_user_selection()

        account_id = (self._choice_index or 0) + self._index_offset
        return address, self._hd_root_path.get_account_path(account_id)

    def _get_user_selection(self) -> str:
        """Prompt the user for a selection."""
        prompt = self._prompt_message
        if self._is_incremented:
            prompt += f"\n\tor 'p' for the previous {self._page_size}"

        # Handle user choice from prompt, including paging.
        return click.prompt(prompt, type=self)

    def _page_from_choice(self, choice):
        choice = choice.lower()
        if choice == "n":
            self._index_offset += self._page_size
            return True
        elif choice == "p" and self._is_incremented:
            self._index_offset -= self._page_size
            return True

        return False

    def _load_choices(self):
        end_range = self._index_offset + self._page_size
        index_range = range(self._index_offset, end_range)
        self.choices = [self._get_address(i) for i in index_range]

    def _get_address(self, account_id: int) -> str:
        path = self._hd_root_path.get_account_path(account_id)
        device = get_device(path)
        return device.get_address()


def get_device(path: "HDAccountPath") -> "LedgerDeviceClient":
    # Perf: lazy load so CLI-usage is faster and abstracted for testing purposes.
    from ape_ledger.client import get_device as _get_device

    return _get_device(path)


__all__ = ["AddressPromptChoice"]



from typing import TYPE_CHECKING, Union

import click
from ape.cli.arguments import existing_alias_argument, non_existing_alias_argument
from ape.cli.options import ape_cli_context, network_option, skip_confirmation_option

if TYPE_CHECKING:
    # NOTE: Type-checking only imports so CLI help loads faster.
    from ape.api import AccountAPI
    from ape_ledger.accounts import LedgerAccount
    from ape_ledger.hdpath import HDAccountPath, HDBasePath

from ape_ledger.exceptions import LedgerSigningError


def _select_account(hd_path: Union["HDBasePath", str]) -> tuple[str, "HDAccountPath"]:
    # NOTE: Lazy import so CLI help loads faster.
    from ape_ledger.choices import AddressPromptChoice

    choices = AddressPromptChoice(hd_path)
    return choices.get_user_selected_account()


@click.group(short_help="Manage Ledger accounts")
def cli():
    """
    Manage Ledger hardware device accounts.
    """


@cli.command("list")
@ape_cli_context()
def _list(cli_ctx):
    """List your Ledger accounts in ape"""

    ledger_accounts = _get_ledger_accounts()

    if len(ledger_accounts) == 0:
        cli_ctx.logger.warning("No accounts found.")
        return

    num_accounts = len(ledger_accounts)
    header = f"Found {num_accounts} account"
    header += "s:" if num_accounts > 1 else ":"
    click.echo(header)

    for account in ledger_accounts:
        alias_display = f" (alias: '{account.alias}')" if account.alias else ""
        hd_path_display = f" (hd-path: '{account.hdpath}')" if account.hdpath else ""
        click.echo(f"  {account.address}{alias_display}{hd_path_display}")


def _get_ledger_accounts() -> list["LedgerAccount"]:
    from ape.utils.basemodel import ManagerAccessMixin

    from ape_ledger.accounts import LedgerAccount

    return [a for a in ManagerAccessMixin.account_manager if isinstance(a, LedgerAccount)]


def _hdpath_callback(ctx, param, val) -> "HDBasePath":
    from ape_ledger.hdpath import HDBasePath

    return HDBasePath(val)


@cli.command()
@ape_cli_context()
@non_existing_alias_argument()
@click.option(
    "--hd-path",
    help=(
        "The Ethereum account derivation path prefix. "
        "Defaults to m/44'/60'/{x}'/0/0 where {{x}} is the account ID. "
        "Exclude {x} to append the account ID to the end of the base path."
    ),
    callback=_hdpath_callback,
)
def add(cli_ctx, alias, hd_path):
    """Add an account from your Ledger hardware wallet"""

    address, account_hd_path = _select_account(hd_path)
    container = cli_ctx.account_manager.containers["ledger"]
    container.save_account(alias, address, str(account_hd_path))
    cli_ctx.logger.success(f"Account '{address}' successfully added with alias '{alias}'.")


def _filter_accounts(acct: "AccountAPI") -> bool:
    from ape_ledger.accounts import LedgerAccount

    return isinstance(acct, LedgerAccount)


@cli.command()
@ape_cli_context()
@existing_alias_argument(account_type=_filter_accounts)
def delete(cli_ctx, alias):
    """Remove a Ledger account from ape"""

    container = cli_ctx.account_manager.containers["ledger"]
    container.delete_account(alias)
    cli_ctx.logger.success(f"Account '{alias}' has been removed.")


@cli.command()
@ape_cli_context()
@skip_confirmation_option("Don't ask for confirmation when removing all accounts")
def delete_all(cli_ctx, skip_confirmation):
    """Remove all Ledger accounts from ape"""

    container = cli_ctx.account_manager.containers["ledger"]
    ledger_accounts = _get_ledger_accounts()
    if len(ledger_accounts) == 0:
        cli_ctx.logger.warning("No accounts found.")
        return

    user_agrees = skip_confirmation or click.confirm("Remove all Ledger accounts from ape?")
    if not user_agrees:
        cli_ctx.logger.info("No account were removed.")
        return

    for account in ledger_accounts:
        container.delete_account(account.alias)
        cli_ctx.logger.success(f"Account '{account.alias}' has been removed.")


@cli.command(short_help="Sign a message with your Ledger device")
@ape_cli_context()
@click.argument("alias")
@click.argument("message", default="Hello World!")
@network_option()
def sign_message(cli_ctx, alias, message, network):
    """Sign a message using a Ledger account"""

    ctx = None
    if network:
        ctx = cli_ctx.network_manager.parse_network_choice(network)
        ctx.__enter__()

    try:
        _sign_message(cli_ctx, alias, message)
    finally:
        if network and ctx and ctx._provider and ctx._provider.is_connected:
            ctx.__exit__()


def _sign_message(cli_ctx, alias, message):
    from eth_account.account import Account
    from eth_account.messages import encode_defunct

    if alias not in cli_ctx.account_manager.aliases:
        cli_ctx.abort(f"Account with alias '{alias}' does not exist.")

    eip191message = encode_defunct(text=message)
    account = cli_ctx.account_manager.load(alias)
    signature = account.sign_message(eip191message)

    if not signature:
        cli_ctx.abort("Failed to sign message.")

    signature_bytes = signature.encode_rsv()

    # Verify signature
    signer = Account.recover_message(eip191message, signature=signature_bytes)
    if signer != account.address:
        cli_ctx.abort(f"Signer resolves incorrectly, got {signer}, expected {account.address}.")

    # Message signed successfully, return signature
    click.echo(signature_bytes.hex())


@cli.command(short_help="Verify a message with your Trezor device")
@ape_cli_context()
@click.argument("message")
@click.argument("signature")
def verify_message(cli_ctx, message, signature):
    from eth_account.account import Account
    from eth_account.messages import encode_defunct

    eip191message = encode_defunct(text=message)

    try:
        signer_address = Account.recover_message(eip191message, signature=signature)
    except ValueError as exc:
        message = "Message cannot be verified. Check the signature and try again."
        raise LedgerSigningError(message) from exc

    alias = (
        cli_ctx.account_manager[signer_address].alias
        if signer_address in cli_ctx.account_manager
        else ""
    )
    click.echo(f"Signer: {signer_address}  {alias}")



from ape.exceptions import AccountsError


class LedgerAccountException(AccountsError):
    """
    An error that occurs in the ape Ledger plugin.
    """


class LedgerSigningError(LedgerAccountException):
    """
    An error that occurs when signing a message or transaction
    using the Ledger plugin.
    """



import json
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Optional

import rich
from ape.api import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, TransactionSignature
from ape_ethereum.transactions import DynamicFeeTransaction, StaticFeeTransaction
from dataclassy import asdict
from eip712 import EIP712Message, EIP712Type
from eth_account.messages import SignableMessage, encode_defunct
from eth_pydantic_types import HexBytes
from eth_utils import is_0x_prefixed, to_bytes

from ape_ledger.client import LedgerDeviceClient, get_device
from ape_ledger.exceptions import LedgerSigningError
from ape_ledger.hdpath import HDAccountPath


def _to_bytes(val) -> bytes:
    if val is None:
        return b""
    elif isinstance(val, str) and is_0x_prefixed(val):
        return to_bytes(hexstr=val)
    elif isinstance(val, str):
        return to_bytes(text=val)
    elif isinstance(val, HexBytes):
        return bytes(val)
    else:
        return to_bytes(val)


class AccountContainer(AccountContainerAPI):
    name: str = "ledger"

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for account_file in self._account_files:
            yield LedgerAccount(container=self, account_file_path=account_file)

    def __setitem__(self, address: AddressType, account: AccountAPI):
        raise NotImplementedError()

    def __delitem__(self, address: AddressType):
        raise NotImplementedError()

    @property
    def _account_files(self) -> Iterator[Path]:
        return self.data_folder.glob("*.json")

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._account_files:
            yield p.stem

    def __len__(self) -> int:
        return len([*self._account_files])

    def save_account(self, alias: str, address: str, hd_path: str):
        """
        Save a new Ledger account to your ape configuration.
        """
        account_data = {"address": address, "hdpath": hd_path}
        path = self.data_folder.joinpath(f"{alias}.json")
        path.write_text(json.dumps(account_data))

    def delete_account(self, alias: str):
        path = self.data_folder.joinpath(f"{alias}.json")
        path.unlink(missing_ok=True)


def _echo_object_to_sign(obj: Any):
    suffix = "Please follow the prompts on your device."
    if isinstance(obj, EIP712Message):

        def make_str(val) -> str:
            if isinstance(val, dict):
                return ", ".join([f"{k}={make_str(v)}" for k, v in val.items()])
            elif isinstance(val, EIP712Type):
                subfields_str = make_str(asdict(val))
                return f"{repr(val)}({subfields_str})"
            elif isinstance(val, (tuple, list, set)):
                inner = ", ".join([make_str(x) for x in val])
                return f"[{inner}]"
            else:
                return f"{val}"

        fields_str = make_str(obj._body_["message"])
        message_str = f"{repr(obj)}({fields_str})"
    else:
        message_str = f"{obj}"

    rich.print(f"{message_str}\n{suffix}")


class LedgerAccount(AccountAPI):
    account_file_path: Path

    @property
    def alias(self) -> str:
        return self.account_file_path.stem

    @property
    def _client(self) -> LedgerDeviceClient:
        return get_device(self.hdpath)

    @property
    def address(self) -> AddressType:
        ecosystem = self.network_manager.get_ecosystem("ethereum")
        return ecosystem.decode_address(self.account_file["address"])

    @property
    def hdpath(self) -> HDAccountPath:
        raw_path = self.account_file["hdpath"]
        return HDAccountPath(raw_path)

    @property
    def account_file(self) -> dict:
        return json.loads(self.account_file_path.read_text())

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        use_eip712_package = isinstance(msg, EIP712Message)
        use_eip712 = use_eip712_package
        if isinstance(msg, str):
            msg_to_sign = encode_defunct(text=msg)
        elif isinstance(msg, int):
            msg_to_sign = encode_defunct(hexstr=HexBytes(msg).hex())
        elif isinstance(msg, bytes):
            msg_to_sign = encode_defunct(primitive=msg)
        elif use_eip712_package:
            # Using eip712 package.
            msg_to_sign = msg.signable_message
        elif isinstance(msg, SignableMessage):
            if msg.version == b"\x01":
                # Using EIP-712 without eip712 package.
                use_eip712 = True
            elif msg.version != b"E":
                try:
                    version_str = msg.version.decode("utf8")
                except Exception:
                    try:
                        version_str = HexBytes(msg.version).hex()
                    except Exception:
                        version_str = "<UnknownVersion>"

                raise LedgerSigningError(
                    f"Unsupported message-signing specification, (version={version_str})."
                )

            msg_to_sign = msg
        else:
            type_name = getattr(type(msg), "__name__", None)
            if not type_name:
                try:
                    type_name = str(type(msg))
                except Exception:
                    type_name = "<UnknownType>"

            raise LedgerSigningError(f"Cannot sign messages of type '{type_name}'.")

        # Echo original message.
        _echo_object_to_sign(msg)

        if use_eip712:
            header = HexBytes(msg_to_sign.header)
            body = HexBytes(msg_to_sign.body)
            signed_msg = self._client.sign_typed_data(header, body)

        else:
            signed_msg = self._client.sign_message(msg_to_sign.body)

        v, r, s = signed_msg
        return MessageSignature(v=v, r=HexBytes(r), s=HexBytes(s))

    def sign_transaction(self, txn: TransactionAPI, **kwargs) -> Optional[TransactionAPI]:
        txn.chain_id = 1
        txn_dict: dict = {
            "nonce": txn.nonce,
            "gas": txn.gas_limit,
            "amount": txn.value,
            "data": _to_bytes(txn.data.hex()),
            "destination": _to_bytes(txn.receiver),
            "chain_id": txn.chain_id,
        }
        if isinstance(txn, StaticFeeTransaction):
            txn_dict["gas_price"] = txn.gas_price

        elif isinstance(txn, DynamicFeeTransaction):
            txn_dict["max_fee_per_gas"] = txn.max_fee
            txn_dict["max_priority_fee_per_gas"] = txn.max_priority_fee
            if txn.access_list:
                txn_dict["access_list"] = [[ls.address, ls.storage_keys] for ls in txn.access_list]

        else:
            raise TypeError(type(txn))

        _echo_object_to_sign(txn)
        v, r, s = self._client.sign_transaction(txn_dict)
        txn.signature = TransactionSignature(
            v=v,
            r=HexBytes(r),
            s=HexBytes(s),
        )
        return txn



import pytest
from ape import accounts
from ape._cli import cli

from ape_ledger.hdpath import HDBasePath


def _get_container():
    return accounts.containers["ledger"]


@pytest.fixture
def alias():
    val = "__integration_test_alias__"
    container = _get_container()
    if val in [a.alias for a in container.accounts]:
        container.delete_account(val)

    return val


@pytest.fixture
def test_account_path(alias):
    return _get_container().data_folder.joinpath(f"{alias}.json")


@pytest.fixture
def existing_account(runner, alias, address, hd_path):
    try:
        container = _get_container()
        container.save_account(alias, address, hd_path)
        yield
    finally:
        _clean_up(runner)


@pytest.fixture(autouse=True)
def clean_after(runner):
    try:
        yield
    finally:
        _clean_up(runner)


@pytest.fixture
def choices(mocker):
    def fn(addr, account_id):
        patch = mocker.patch("ape_ledger._cli._select_account")

        def se(hd_path):
            return addr, HDBasePath(hd_path).get_account_path(account_id)

        patch.side_effect = se

    return fn


def _clean_up(runner):
    runner.invoke(cli, ("ledger", "delete", alias), input="y")


def _get_account_path(alias=alias):
    container = _get_container()
    return container.data_folder.joinpath(f"{alias}.json")


@pytest.mark.parametrize("cmd", (["ledger", "list"], ["accounts", "list", "--all"]))
def test_list(runner, existing_account, cmd, address, alias):
    result = runner.invoke(cli, cmd)
    assert result.exit_code == 0, result.output
    assert alias in result.output
    assert address.lower() in result.output.lower()


def test_add(runner, assert_account, address, alias, choices, hd_path):
    container = _get_container()
    choices(address, 2)
    result = runner.invoke(cli, ("ledger", "add", alias), catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert f"Account '{address}' successfully added with alias '{alias}'." in result.output

    expected_path = container.data_folder.joinpath(f"{alias}.json")
    expected_hd_path = "m/44'/60'/2'/0/0"
    assert_account(expected_path, expected_hdpath=expected_hd_path)


def test_add_when_hd_path_specified(runner, alias, address, hd_path, assert_account, choices):
    test_hd_path = "m/44'/60'/0'"
    container = _get_container()
    choices(address, 2)
    result = runner.invoke(
        cli,
        ("ledger", "add", alias, "--hd-path", test_hd_path),
    )
    assert result.exit_code == 0, result.output
    assert f"Account '{address}' successfully added with alias '{alias}'." in result.output

    expected_path = container.data_folder.joinpath(f"{alias}.json")
    expected_hd_path = "m/44'/60'/0'/2"
    assert_account(expected_path, expected_hdpath=expected_hd_path)


def test_add_alias_already_exists(runner, existing_account, choices, address, alias):
    choices(address, 2)

    # Ensure exists
    runner.invoke(cli, ("ledger", "add", alias))

    result = runner.invoke(cli, ("ledger", "add", alias))
    assert result.exit_code == 1, result.output
    assert "ERROR:" in result.output
    assert "(AliasAlreadyInUseError)" in result.output
    assert f"Account with alias '{alias}' already in use." in result.output


def test_delete(runner, existing_account, alias):
    result = runner.invoke(cli, ("ledger", "delete", alias))
    assert result.exit_code == 0, result.output
    assert f"Account '{alias}' has been removed" in result.output


def test_delete_account_not_exists(runner, alias):
    not_alias = f"{alias}TYPO"
    result = runner.invoke(cli, ("ledger", "delete", not_alias))
    assert result.exit_code == 2
    assert f"'{not_alias}'" in result.output



import pytest

from ape_ledger.choices import AddressPromptChoice


@pytest.fixture(autouse=True)
def patch_device(device_factory):
    return device_factory("choices")


class TestAddressPromptChoice:
    def test_get_user_selected_account(self, mocker, hd_path, address):
        mock_prompt = mocker.patch("ape_ledger.choices.click.prompt")
        choices = AddressPromptChoice(hd_path)
        choices._choice_index = 3
        choices._index_offset = 2

        # `None` means the user hasn't selected yeet
        # And is entering other keys, possible the paging keys.
        mock_prompt_return_values = iter((None, None, None, None, address, None))

        def _side_effect(*args, **kwargs):
            return next(mock_prompt_return_values)

        mock_prompt.side_effect = _side_effect
        address, hdpath = choices.get_user_selected_account()
        assert address == address
        assert str(hdpath) == f"m/44'/60'/{choices._choice_index + choices._index_offset}'/0/0"



import atexit
from functools import cached_property
from typing import TYPE_CHECKING

import hid  # type: ignore
from ape.logging import LogLevel, logger
from ledgerblue.comm import HIDDongleHIDAPI, getDongle  # type: ignore
from ledgereth.accounts import get_account_by_path
from ledgereth.messages import sign_message, sign_typed_data_draft
from ledgereth.transactions import SignedType2Transaction, create_transaction

if TYPE_CHECKING:
    from ape_ledger.hdpath import HDAccountPath


class DeviceFactory:
    device_map: dict[str, "LedgerDeviceClient"] = {}

    def create_device(self, account: "HDAccountPath"):
        if account.path in self.device_map:
            return self.device_map[account.path]

        device = LedgerDeviceClient(account)
        self.device_map[account.path] = device
        return device


def get_dongle(debug: bool = False, reopen_on_fail: bool = True) -> HIDDongleHIDAPI:
    try:
        return getDongle(debug=debug)
    except (OSError, RuntimeError) as err:
        if str(err).lower().strip() in ("open failed", "already open") and reopen_on_fail:
            # Device was not closed properly.
            device = hid.device()
            device.close()
            return get_dongle(debug=debug, reopen_on_fail=False)

        raise  # the OSError


class LedgerDeviceClient:
    def __init__(self, account: "HDAccountPath"):
        self._account = account.path.lstrip("m/")

    @cached_property
    def dongle(self):
        debug = logger.level <= LogLevel.DEBUG
        device = get_dongle(debug=debug)

        def close():
            logger.info("Closing device.")
            device.close()

        atexit.register(close)
        return device

    def get_address(self) -> str:
        return get_account_by_path(self._account, dongle=self.dongle).address

    def sign_message(self, text: bytes) -> tuple[int, int, int]:
        signed_msg = sign_message(text, sender_path=self._account, dongle=self.dongle)
        return signed_msg.v, signed_msg.r, signed_msg.s

    def sign_typed_data(self, domain_hash: bytes, message_hash: bytes) -> tuple[int, int, int]:
        signed_msg = sign_typed_data_draft(
            domain_hash, message_hash, sender_path=self._account, dongle=self.dongle
        )
        return signed_msg.v, signed_msg.r, signed_msg.s

    def sign_transaction(self, txn: dict) -> tuple[int, int, int]:
        kwargs = {**txn, "sender_path": self._account, "dongle": self.dongle}
        signed_tx = create_transaction(**kwargs)
        return (
            (signed_tx.y_parity, signed_tx.sender_r, signed_tx.sender_s)
            if isinstance(signed_tx, SignedType2Transaction)
            else (signed_tx.v, signed_tx.r, signed_tx.s)
        )


_device_factory = DeviceFactory()


def get_device(account: "HDAccountPath") -> LedgerDeviceClient:
    return _device_factory.create_device(account)



import pytest

from ape_ledger.hdpath import HDAccountPath, HDBasePath, HDPath


class TestHDPath:
    include_subclasses = pytest.mark.parametrize("cls", (HDPath, HDBasePath, HDAccountPath))

    @include_subclasses
    def test_init_no_m_raises_value_errors(self, cls):
        path_without_m = "/44'/60'/0'/0"
        with pytest.raises(ValueError) as err:
            cls(path_without_m)

        assert str(err.value) == "HD path must begin with m/"

    @include_subclasses
    def test_account_path_trailing_slash_removes_trailing_slash(self, cls):
        path_with_trailing_slash = "m/44'/60'/0'/0/"
        path_obj = cls(path_with_trailing_slash)
        assert path_obj.path[-1] != "/"


class TestHDBasePath:
    def test_init_x_not_in_path_and_trailing_slash(self):
        # Edge case since the trailing slash is handled in super class
        # and when there is no {x} in the path, we append it to the end
        path_without_x_and_trailing_slash = "m/44'/60'/0'/0/"
        obj = HDBasePath(path_without_x_and_trailing_slash)
        actual = obj.path
        expected = "m/44'/60'/0'/0/{x}"
        assert actual == expected

    def test_get_account_path_x_in_path(self):
        path_with_x = "m/44'/60'/{x}'/0/0"
        obj = HDBasePath(path_with_x)
        actual = obj.get_account_path(3).path
        expected = "m/44'/60'/3'/0/0"
        assert actual == expected

    def test_get_account_path_x_not_in_path(self):
        path_without_x = "m/44'/60'/0'/0"
        obj = HDBasePath(path_without_x)
        actual = obj.get_account_path(3).path
        expected = "m/44'/60'/0'/0/3"
        assert actual == expected


class TestHDAccountPath:
    def test_as_bytes(self):
        path = HDAccountPath("m/44'/60'/2'/0/0")
        actual = path.as_bytes()
        expected = b"\x05\x80\x00\x00,\x80\x00\x00<\x80\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00"
        assert actual == expected



from importlib import import_module
from typing import Any

from ape import plugins


@plugins.register(plugins.AccountPlugin)
def account_types():
    from ape_ledger.accounts import AccountContainer, LedgerAccount

    return AccountContainer, LedgerAccount


def __getattr__(name: str) -> Any:
    if name in ("AccountContainer", "LedgerAccount"):
        return getattr(import_module("ape_ledger.accounts"), name)

    else:
        raise AttributeError(name)


__all__ = [
    "AccountContainer",
    "LedgerAccount",
]
